//
// Zhadum.h
//
// Copyright (c) 2001-2002 Mark H. P. Lord. All rights reserved.
// 
// This software is provided 'as-is', without any express or implied
// warranty.  In no event will the author(s) be held liable for any damages
// arising from the use of this software.
// 
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
// 
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would be
//    appreciated but is not required.
// 2. Altered source versions must be plainly marked as such, and must not be
//    misrepresented as being the original software.
// 3. This notice may not be removed or altered from any source distribution.
//

#ifndef _ZHADUM_H_
#define _ZHADUM_H_

#ifndef __WINDOWS__
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

// Begins automatically applying window shadows to our popup windows
BOOL WINAPI ZhadumBegin (HINSTANCE hinst);

// Exclude the specified window from window shadows. Returns FALSE if there
// are too many exclusions in place.
BOOL WINAPI ZhadumExcludeWindow (HWND hwnd);

// Remove the exclusion on the specified window. This won't create shadows
// but it will allow any future windows with the same HWND to have shadows.
// You no longer have to call this, it's handled internally.
void WINAPI ZhadumUnexcludeWindow (HWND hwnd);

// Exclude the specified WNDCLASS (can be an internal class, e.g. WC_DIALOG)
// Returns FALSE if there are too many exclusions in place.
BOOL WINAPI ZhadumExcludeClass (LPCTSTR classname);

// Set the alpha of the shadows belonging to the specified window.
void WINAPI ZhadumSetWindowShadowAlpha (HWND hwnd, BYTE alpha);

// Stop auto-shadowing
void WINAPI ZhadumEnd ();

// Display a Win32 error message in a MessageBox
void WINAPI ZhadumDisplayWinError (DWORD err, LPCTSTR title);

#ifdef __cplusplus
}
#endif

#endif
